;
; source code for FMan's Game of Life by John Horton Conway in 1970
;
; 2010 by Sami Klemola. All rights reserved.
;

	!to "\bin\vice\files\FMan-GameLife.prg",cbm
	*=$801

	firstlo = 92
	firsthi = 93
	secondlo = 94
	secondhi = 95
	thirdlo = 96
	thirdhi = 97
	destlo = 98
	desthi = 99
	row = 100
	live = 101
	temp = 102
	edit = 103

	; BASIC line to start the MC

	!byte $b,8,0,0,$9e,$32,$30,$36,$31,0,0,0

	lda #7
	sta $286
	lda #<string
	ldx #>string
	jsr print
	lda #<NMI		; non-maskable interrupt for breaking program
	sta $318
	lda #>NMI
	sta $319
	lda #127			; disable CIA2 interrupts except Restore
	sta $dd0d
	lda #32			; set screen position to $2000
	sta $d2
	sta $288
rst	cli			; re-enable maskable interrupts for keyboard
	lda #21
	sta $d018		; show normal screen for menu
	sta edit
	lda #<loopd
	sta Finnish-2
	lda #>loopd
	sta Finnish-1		; restore loop address for the jump (see below)
	lda #0
	sta 53280
	sta 53281
	tax
cls	sta $2000,x		; clear the two screens used
	sta $2100,x
	sta $2200,x
	sta $2300,x
	sta $2400,x
	sta $2500,x
	sta $2600,x
	sta $2700,x
	inx
	bne cls

key	jsr $ffe4		; read from keyboard
	beq key
	sec
	sbc #48			; values 0-9 represent keys 0-9
	bne next
	sta edit			; Make your own (key 0)
	beq justdoit
next	cmp #10			; try again if not 1-9
	bcs key
	tax			; use keypress as index to pattern data table
	ldy #0
	lda #<(patterns-4)
	sta destlo
	lda #>(patterns-4)
	sta desthi
loopi	lda (destlo),y		; look for double-zero end marker
	bne nonzero
	jsr incdest
	lda (destlo),y
	beq found
nonzero	jsr incdest
	bne loopi		; unconditional

incdest	inc destlo		; increment data pointer
	bne return
	inc desthi
return	rts

found	jsr incdest
	dex			; traverse table until the correct entry
	bne loopi
loopj	lda (destlo),y		; get address from data table
	sta store+1
	jsr incdest
	lda (destlo),y
	sta store+2
	beq zero
loopk	jsr incdest
store	dec $2000		; turn on the target cell (0 to 255)
	bne loopj
zero	lda store+1
	bne loopk		; only LSB was zero, so just go on

justdoit	lda #%111100		; char
	sta $27fa
	sta $27fb
	sta $27fc
	sta $27fd
	lda edit
	bne goahead
	jsr $e544
foobar	ldy #0
	sty destlo
	lda #32
	sta desthi
loopl	lda (destlo),y		; change the 0-bytes into spaces
	beq loopn
	cmp #32
	beq loopm
	lda #81			; change live cells to <CTRL>-Q
	bne loopo
loopn	lda #32
loopo	sta (destlo),y
loopm	iny
	bne loopl
	inc desthi
	lda desthi
	cmp #36
	bcc loopl
	lda #$85			; screen one with normal font
	sta $d018
looph	jsr $ffcf		; invoke screen editor
	cmp #13
	bne looph		; read input until CR appears
	lda #32
	sta desthi
loopf	lda (destlo),y		; process the screen into 0/255
	beq foobar3
	cmp #32
	beq foobar1
	lda #255
	bne foobar2
foobar1	lda #0
foobar2	sta (destlo),y
foobar3	iny
	bne loopf
	inc desthi
	lda desthi
	cmp #36
	bcc loopf

goahead	lda #$88			; show initial screen
	sta $d018
wait	jsr $ffe4		; await keypress
	beq wait
	cmp #69			; go to editor if 'E'
	beq foobar
	sei
loopc	ldx #32			; main loop
	ldy #36
	lda #$98			; show screen two
	jsr iterate
	ldx #36
	ldy #32
	lda #$88			; show screen one
	jsr iterate
	bne loopc

iterate	pha			; do next generation
	sty desthi
	ldy #0
	sty row
	sty firstlo
	stx firsthi
	stx secondhi
	stx thirdhi
	lda #40
	sta secondlo
	sta destlo
	lda #80
	sta thirdlo
	iny			; start from column 1
loopd	dey
	ldx #0			; use X to count neighbors
	lda (firstlo),y
	beq testa
	inx
testa	lda (secondlo),y
	beq testb
	inx
testb	lda (thirdlo),y
	beq testc
	inx
testc	iny
	lda (firstlo),y
	beq testd
	inx
testd	lda (secondlo),y
	sta live			; current cell's state this generation
	lda (thirdlo),y
	beq teste
	inx
teste	iny
	lda (firstlo),y
	beq testf
	inx
testf	lda (secondlo),y
	beq testg
	inx
testg	lda (thirdlo),y
	beq testh
	inx
testh	lda live			; was this cell alive now?
	beq nope
	cpx #2			; check its neighborcount
	bcs testi

;   1. Any live cell with fewer than two live neighbours dies, as if caused by underpopulation.
;   2. Any live cell with more than three live neighbours dies, as if by overcrowding.
;   3. Any live cell with two or three live neighbours lives on to the next generation.
;   4. Any dead cell with exactly three live neighbours becomes a live cell.

die	lda #0			; less than two, so death
	beq done
testi	cpx #4			; more than three, so death
	bcs die
	bcc alive
nope	cpx #3			; dead cell with 3 neighbors?
	bne die
alive	lda #255			; live cell (2-3 live neighbors)	
done	dey
	sta (destlo),y		; write the destination
	iny
	cpy #39			; run it through columns 1-38
	bcc loopd
	lda secondlo		; calculate addresses of next line
	sta firstlo
	lda secondhi
	sta firsthi
	lda thirdhi
	cmp secondhi
	beq foo
	inc desthi
foo	sta secondhi
	lda thirdlo
	sta secondlo
	sta destlo
	clc
	adc #40
	sta thirdlo
	bcc bar
	inc thirdhi
bar	inc row			; loop whole screen
	lda row
	cmp #23
	bcs Finnish
	ldy #1
	jmp loopd		; too far for a short branch (address is overwritten)
Finnish	lda #242
loopa	cmp $d012		; wait for bottom border before changing the screen
	bne loopa
	pla
	sta $d018
	rts

print	ldy #0			; print a null-terminated string (A,X)
	sta destlo
	stx desthi
printa	lda (destlo),y
	beq printb
	jsr $ffd2
	iny
	bne printa
printb	rts

NMI	lda #<rst		; change the loop jump with the restart address
	sta Finnish-2
	lda #>rst		; this will break the loop and cause a jump to the
	sta Finnish-1		; restart address, so destroying Akku won't matter
	rti			; and we just return to wait for the jump

string	!pet $93,13,"game of life by fman",13,"--------------------",13,13,13,"select pattern",13,13,"1  glider",13,"2  lightweight spaceship",13,"3  pentadecathlon",13,"4  glider synthesis",13,"5  die hard",13,"6  acorn",13,"7  90",13,"8  gosper glider gun",13,"9  f-pentomino",13,"0  make your own",13,13,13,"press a numeric key",0,0 ; second null for fake non-existing zero pattern end

patterns	!word $2102-88-81,$2102-88-40,$2100-88,$2101-88,$2102-88,0 ; Glider
	!word $2282,$2282+3,$2282+44,$2282+80,$2282+84,$2282+121,$2282+122,$2282+123,$2282+124,0 ; LWSS
	!word $2101+158+2,$2101+158+3,$2101+158+4,$2101+158+5,$2101+158+6,$2101+158+7,$2101+158+41,$2101+158+41+7,$2101+158+80,$2101+158+80+9,$2101+158+121,$2101+158+121+7,$2101+158+162,$2101+158+163,$2101+158+164,$2101+158+165,$2101+158+166,$2101+158+167,0 ; Pentadecathlon
	!word $2101+75+6-156,$2101+75+46-156,$2101+75+48-156,$2101+75+86-156,$2101+75+87-156,$2101+75+160+156,$2101+75+161+156,$2101+75+162+156,$2101+75+202+156,$2101+75+241+156,$2101+75+247+156,$2101+75+248+156,$2101+75+288+156,$2101+75+289+156,$2101+75+327+156,0 ; Glider synthesis
	!word $2100+46,$2100+80,$2101+80,$2101+80+40,$2101+80+44,$2101+80+45,$2101+80+46,0 ; Diehard
	!word $2100-79+400,$2100-37+400,$2100+400,$2101+400,$2104+400,$2105+400,$2106+400,0 ; Acorn
	!word $2101+84+160,$2101+84+161,$2101+84+162,$2101+84+160+40,$2101+84+160+42,$2101+84+160+80,$2101+84+160+81,$2101+84+160+82,$2101+84+160+122,$2101+84+160+160,$2101+84+160+161,$2101+84+160+162,$2101+84+164,$2101+84+165,$2101+84+166,$2101+84+164+40,$2101+84+164+42,$2101+84+164+80,$2101+84+164+82,$2101+84+164+120,$2101+84+164+122,$2101+84+164+160,$2101+84+164+161,$2101+84+164+162,0 ; 90
	!word $207a+24,$207a+62,$207a+64,$207a+92,$207a+93,$207a+100,$207a+101,$207a+114,$207a+115,$207a+131,$207a+135,$207a+140,$207a+141,$207a+154,$207a+155,$207a+160,$207a+161,$207a+170,$207a+176,$207a+180,$207a+181,$207a+200,$207a+201,$207a+210,$207a+214,$207a+216,$207a+217,$207a+222,$207a+224,$207a+250,$207a+256,$207a+264,$207a+291,$207a+295,$207a+332,$207a+333,0 ; Gosper glider gun
	!word $21a9+200,$21aa+200,$21a8+200+40,$21a8+200+41,$21a8+200+81,0 ; F-pentomino
